import sys

import click

from clipped.formatting import Printer
from clipped.utils import indentation
from clipped.utils.dicts import dict_to_tabulate
from clipped.utils.versions import clean_version_for_check, compare_versions

from polyaxon import pkg
from polyaxon._cli.session import set_versions_config
from polyaxon._deploy.operators.pip import PipOperator
from polyaxon.logger import clean_outputs, logger

PROJECT_CLI_NAME = "polyaxon"


def pip_upgrade(project_name=PROJECT_CLI_NAME):
    PipOperator.execute(["install", "--upgrade", project_name], stream=True)
    Printer.print("polyaxon upgraded!")


def get_version(package: str, show_error: bool = True):
    import pkg_resources

    try:
        return pkg_resources.get_distribution(package).version
    except pkg_resources.DistributionNotFound:
        if show_error:
            logger.error("`%s` is not installed", package)


def check_old_packages():
    pkg = "polyaxon-cli"
    if get_version(package=pkg, show_error=False):
        Printer.error(
            "Legacy package `{pkg}` is installed. Please run `pip uninstall {pkg}`".format(
                pkg=pkg
            ),
            sys_exit=True,
        )


def get_current_version():
    return pkg.VERSION


def check_cli_version(config, is_cli: bool = True):
    """Check if the current cli version satisfies the server requirements"""

    min_version = clean_version_for_check(config.min_version)
    latest_version = clean_version_for_check(config.latest_version)
    current_version = clean_version_for_check(config.current_version)
    if not min_version or not latest_version or not current_version:
        if is_cli:
            Printer.error(
                "Could not get the min/latest versions from compatibility API.",
                sys_exit=True,
            )
        else:
            return
    if compare_versions(current=current_version, reference=min_version, comparator="<"):
        Printer.print(
            "Your version of Polyaxon CLI ({}) is no longer supported.".format(
                config.current_version
            )
        )
        if click.confirm(
            "Do you want to upgrade to " "version {} now?".format(config.latest_version)
        ):
            pip_upgrade()
            sys.exit(0)
        else:
            indentation.puts(
                "To upgrade to the latest version `{}`, "
                "you can manually run:".format(config.latest_version)
            )
            with indentation.indent(4):
                indentation.puts("pip install -U polyaxon")

            sys.exit(0)
    elif compare_versions(
        current=current_version, reference=latest_version, comparator="<"
    ):
        indentation.puts(
            "New version of Polyaxon CLI ({}) is now available. To upgrade run:".format(
                config.latest_version
            )
        )
        with indentation.indent(4):
            indentation.puts("pip install -U polyaxon")
    elif compare_versions(
        current=current_version, reference=latest_version, comparator=">"
    ):
        indentation.puts(
            "Your version of CLI ({}) is ahead of the latest version "
            "supported by Polyaxon server ({}) on your cluster, "
            "and might be incompatible.".format(
                config.current_version, config.latest_version
            )
        )


@click.command()
@click.option(
    "--check", is_flag=True, default=False, help="Check compatibility versions."
)
@clean_outputs
def version(check):
    """Print the current version of the cli and platform."""
    check_old_packages()
    Printer.heading("Current cli version: {}".format(pkg.VERSION))
    if check:
        config = set_versions_config()
        Printer.heading("Platform version:")
        config_installation = (
            dict_to_tabulate(
                config.installation.to_dict(),
                humanize_values=True,
                exclude_attrs=["hmac", "auth", "host"],
            )
            if config.installation
            else {"Server": "not found or not deployed"}
        )
        Printer.dict_tabulate(config_installation)
        Printer.heading("Compatibility versions:")
        Printer.dict_tabulate(
            config.compatibility.to_dict() if config.compatibility else {}
        )
        check_cli_version(config)


@click.command()
@clean_outputs
def upgrade():
    """Install/Upgrade polyaxon cli."""
    try:
        pip_upgrade(PROJECT_CLI_NAME)
    except Exception as e:
        logger.error(e)



from typing import Dict, Optional, Union

from clipped.compact.pydantic import Field, StrictStr, validator
from clipped.utils.versions import clean_version_post_suffix

from polyaxon import pkg
from polyaxon._containers.pull_policy import PullPolicy
from polyaxon._k8s import k8s_schemas, k8s_validation
from polyaxon._schemas.base import BaseSchemaModel


def get_init_resources() -> k8s_schemas.V1ResourceRequirements:
    return k8s_schemas.V1ResourceRequirements(
        limits={"cpu": "1", "memory": "500Mi"},
        requests={"cpu": "0.1", "memory": "60Mi"},
    )


class V1PolyaxonInitContainer(BaseSchemaModel):
    """Polyaxon init is a helper container that initialize the environment
    required for the main container to function correctly.

    Polyaxon CE and Polyaxon Agent are deployed with default values for the init container,
    however if you need to control or update one or several aspects
    of how the init container that gets injected, this guide walks through the possible options.

    Args:
        image: str, optional.
        image_tag: str, optional.
        image_pull_policy: str, optional.
        resources: V1ResourceRequirements, optional.

    ## YAML usage

    ```yaml
    >>> init:
    >>>   image: polyaxon/polyaxon-sidecar
    >>>   imageTag: v1.x
    >>>   imagePullPolicy: IfNotPresent
    >>>   resources:
    >>>     requests:
    >>>       memory: "64Mi"
    >>>       cpu: "50m"
    ```

    ## Fields

    ### image

    The container image to use.

    ```yaml
    >>> init:
    >>>   image: polyaxon/polyaxon-sidecar
    ```

    ### imageTag

    The container image tag to use.

    ```yaml
    >>> init:
    >>>   imageTag: dev
    ```

    ### imagePullPolicy

    The image pull policy to use, it must be a valid policy supported by Kubernetes.

    ```yaml
    >>> init:
    >>>   imagePullPolicy: Always
    ```

    ### resources

    The resources requirements to allocate to the container.

    ```yaml
    >>> init:
    >>>   resources:
    >>>     memory: "64Mi"
    >>>     cpu: "50m"
    ```

    > **N.B.1**: Resources are applied to all instances of
    > the init container within the same pod.

    > **N.B.2**: It's possible to alter this behaviour on per operation level
    > using the init section.
    """

    _IDENTIFIER = "container"
    _SWAGGER_FIELDS = ["resources"]

    image: Optional[StrictStr]
    image_tag: Optional[StrictStr] = Field(alias="imageTag")
    image_pull_policy: Optional[PullPolicy] = Field(alias="imagePullPolicy")
    resources: Optional[Union[k8s_schemas.V1ResourceRequirements, Dict]]

    @validator("resources", always=True, pre=True)
    def validate_resources(cls, v) -> k8s_schemas.V1ResourceRequirements:
        return k8s_validation.validate_k8s_resource_requirements(v)

    def get_image(self) -> str:
        image = self.image or "polyaxon/polyaxon-init"
        image_tag = (
            self.image_tag
            if self.image_tag is not None
            else clean_version_post_suffix(pkg.VERSION)
        )
        return "{}:{}".format(image, image_tag) if image_tag else image

    def get_resources(self) -> k8s_schemas.V1ResourceRequirements:
        return self.resources if self.resources else get_init_resources()


def get_default_init_container(
    schema: bool = True,
) -> Union[Dict, V1PolyaxonInitContainer]:
    default = {
        "image": "polyaxon/polyaxon-init",
        "imageTag": clean_version_post_suffix(pkg.VERSION),
        "imagePullPolicy": PullPolicy.IF_NOT_PRESENT.value,
        "resources": {
            "limits": {"cpu": "1", "memory": "500Mi"},
            "requests": {"cpu": "0.1", "memory": "60Mi"},
        },
    }
    if schema:
        return V1PolyaxonInitContainer.from_dict(default)
    return default



import sys

import click

from clipped.formatting import Printer
from clipped.utils.dicts import dict_to_tabulate
from clipped.utils.paths import check_or_create_path

from polyaxon import settings
from polyaxon._cli.errors import handle_cli_error
from polyaxon._cli.session import set_versions_config
from polyaxon._managers.cli import CliConfigManager
from polyaxon._managers.client import ClientConfigManager
from polyaxon._managers.home import HomeConfigManager
from polyaxon._managers.user import UserConfigManager
from polyaxon.logger import clean_outputs, logger


def set_home_path(home_path: str):
    try:
        _config = HomeConfigManager.get_config_or_default()
    except Exception as e:
        logger.debug(
            "Home configuration could not be loaded.\n"
            "Error: %s\n"
            "Purging home configuration and resetting values.",
            e,
        )
        logger.debug()
        HomeConfigManager.purge()
        _config = HomeConfigManager.get_config_or_default()
    try:
        check_or_create_path(home_path, is_dir=False)
    except Exception as e:
        handle_cli_error(
            e, message=f"Couldn't create path configuration at {home_path}."
        )
        Printer.heading(
            "Please make sure that that the path is accessible or manually create it."
        )
        sys.exit(1)
    setattr(_config, "path", home_path)
    HomeConfigManager.set_config(_config)
    settings.set_home_config(_config)


def validate_options(ctx, param, value):
    possible_values = ["verbose", "host"]
    if value and value not in possible_values:
        raise click.BadParameter(
            "Value `{}` is not supported, must one of the value {}".format(
                value, possible_values
            )
        )
    return value


@click.group(invoke_without_command=True)
@click.option(
    "--list",
    "-l",
    "_list",
    is_flag=True,
    help="Deprecated, please use `polyaxon config show`.",
)
@clean_outputs
def config(_list):  # pylint:disable=redefined-builtin
    """Set and get the global configurations."""
    if _list:
        Printer.warning(
            "`polyaxon config -l` is deprecated, please use `polyaxon config show`!"
        )


@config.command()
@clean_outputs
def show():
    """Show the current cli, client, and user configs."""
    _config = HomeConfigManager.get_config_or_default()
    Printer.heading(
        "In addition to environment variables, global configs will be loaded from:"
    )
    Printer.dict_tabulate(_config.to_dict())
    _config = ClientConfigManager.get_config_or_default()
    Printer.heading("Client config:")
    Printer.dict_tabulate(_config.to_dict())
    _config = CliConfigManager.get_config_or_default()
    if _config:
        Printer.heading("CLI config:")
        if _config.current_version:
            Printer.print("Version {}".format(_config.current_version))
        else:
            Printer.warning("This cli is not configured.")
        if _config.installation:
            config_installation = dict_to_tabulate(
                _config.installation,
                humanize_values=True,
                exclude_attrs=["hmac", "auth", "host"],
            )
            Printer.heading("Platform config:")
            Printer.dict_tabulate(config_installation)
        else:
            Printer.warning("This cli is not connected to a Polyaxon Host.")
    _config = UserConfigManager.get_config_or_default()
    if _config:
        Printer.heading("User config:")
        config_user = dict_to_tabulate(
            _config,
            humanize_values=True,
            exclude_attrs=["theme"],
        )
        Printer.dict_tabulate(config_user)


@config.command()
@click.argument("keys", type=str, nargs=-1)
@clean_outputs
def get(keys):
    """Get the specific keys from the global configuration.

    Examples:

    \b
    $ polyaxon config get home host verify-ssl
    """

    if not keys:
        return

    keys = set(keys)

    # Check Home config
    _config = HomeConfigManager.get_config_or_default()
    print_values = {}
    if "home" in keys:
        if hasattr(_config, "path"):
            print_values["home"] = getattr(_config, "path")
            keys.discard("home")
    if print_values:
        Printer.heading("Home config:")
        Printer.dict_tabulate(print_values)

    not_found_keys = set([])

    # Check client config
    _config = ClientConfigManager.get_config_or_default()
    print_values = {}
    for key in keys:
        key = key.replace("-", "_")
        if hasattr(_config, key):
            print_values[key] = getattr(_config, key)
            not_found_keys.discard(key)
        else:
            not_found_keys.add(key)
    if print_values:
        Printer.heading("Client config:")
        Printer.dict_tabulate(print_values)

    keys = not_found_keys
    not_found_keys = set([])

    # Check cli config
    _config = CliConfigManager.get_config_or_default()
    print_values = {}
    for key in keys:
        key = key.replace("-", "_")
        if hasattr(_config, key):
            print_values[key] = getattr(_config, key)
            not_found_keys.discard(key)
        else:
            not_found_keys.add(key)
    if print_values:
        Printer.heading("CLI config:")
        Printer.dict_tabulate(print_values)

    if not_found_keys:
        Printer.print(
            "The following keys `{}` were not found in any configuration.".format(
                not_found_keys
            )
        )


@config.command(name="set")
@click.option("--debug", type=bool, help="To set the verbosity of the client.")
@click.option("--host", type=str, help="To set the server endpoint.")
@click.option(
    "--no-api",
    type=bool,
    help="To disable any API call.",
)
@click.option(
    "--verify-ssl",
    type=bool,
    help="To set whether or not to verify the SSL certificate.",
)
@click.option(
    "--home",
    type=click.Path(exists=False),
    help="To set POLYAXON_HOME to specify the context where the CLI/Client reads/writes global configuration.",
)
@click.option(
    "--disable-errors-reporting",
    type=bool,
    help="To set the disable errors reporting.",
)
@click.option(
    "--no-purge",
    is_flag=True,
    default=False,
    help="To reconfigure the host without purging auth and other config options.",
)
@clean_outputs
def set_(**kwargs):  # pylint:disable=redefined-builtin
    """Set the global config values.

    Examples:

    \b
    $ polyaxon config set --home=/tmp/.polyaxon

    \b
    $ polyaxon config set --home=

    \b
    $ polyaxon config set --host=localhost
    """
    no_purge = kwargs.pop("no_purge", None)
    if kwargs.get("home") is not None:
        home_path = kwargs.pop("home", None)
        set_home_path(home_path)

    from polyaxon._managers.auth import AuthConfigManager

    try:
        _config = ClientConfigManager.get_config_or_default()
    except Exception as e:
        handle_cli_error(e, message="Polyaxon load configuration.")
        Printer.heading("You can reset your config by running: `polyaxon config purge`")
        sys.exit(1)

    should_purge = False
    for key, value in kwargs.items():
        if key == "host" and value is not None:
            should_purge = True
        if value is not None:
            setattr(_config, key, value)

    ClientConfigManager.set_config(_config)
    Printer.success("Config was updated.")
    # Reset cli config
    CliConfigManager.purge()
    if should_purge and not no_purge:
        AuthConfigManager.purge()
        UserConfigManager.purge()
        settings.set_client_config()
        set_versions_config()


@config.command()
@click.option(
    "--cache-only",
    is_flag=True,
    help="To purge the cache only.",
)
@clean_outputs
def purge(cache_only):
    """Purge the global config values."""
    from polyaxon._managers.auth import AuthConfigManager
    from polyaxon._managers.project import ProjectConfigManager
    from polyaxon._managers.run import RunConfigManager

    if not cache_only:
        ClientConfigManager.purge()
        CliConfigManager.purge()
        AuthConfigManager.purge()
        UserConfigManager.purge()
    ProjectConfigManager.purge()
    RunConfigManager.purge()
    Printer.success("Configs was removed.")



import datetime

from typing import List, Optional, Tuple

from clipped.utils.tz import now
from kubernetes_asyncio.client.models import V1Pod
from kubernetes_asyncio.client.rest import ApiException

from polyaxon._flow import V1RunKind
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from traceml.logging import V1Log, V1Logs


async def handle_container_logs(
    k8s_manager: AsyncK8sManager, pod: V1Pod, container_name: str, **params
) -> List[V1Log]:
    resp = None
    try:
        resp = await k8s_manager.k8s_api.read_namespaced_pod_log(
            pod.metadata.name,
            k8s_manager.namespace,
            container=container_name,
            timestamps=True,
            **params,
        )
    except ApiException:
        pass
    if not resp:
        return []

    logs = []
    for log_line in resp.split("\n"):
        if log_line:
            logs.append(
                V1Log.process_log_line(
                    value=log_line,
                    node=pod.spec.node_name,
                    pod=pod.metadata.name,
                    container=container_name,
                )
            )
    return logs


async def handle_pod_logs(
    k8s_manager: AsyncK8sManager, pod: V1Pod, **params
) -> List[V1Log]:
    logs = []
    for container in pod.spec.init_containers or []:
        logs += await handle_container_logs(
            k8s_manager=k8s_manager, pod=pod, container_name=container.name, **params
        )
    for container in pod.spec.containers or []:
        logs += await handle_container_logs(
            k8s_manager=k8s_manager, pod=pod, container_name=container.name, **params
        )
    return logs


async def query_k8s_operation_logs(
    k8s_manager: AsyncK8sManager,
    instance: str,
    last_time: Optional[datetime.datetime],
    stream: bool = False,
) -> Tuple[List[V1Log], Optional[datetime.datetime]]:
    new_time = now()
    params = {}
    if last_time:
        since_seconds = (new_time - last_time).total_seconds() - 1
        params["since_seconds"] = int(since_seconds)
    if stream:
        params["tail_lines"] = V1Logs._CHUNK_SIZE
    logs = []

    pods = await k8s_manager.list_pods(
        label_selector=k8s_manager.get_managed_by_polyaxon(instance)
    )

    for pod in pods:
        logs += await handle_pod_logs(
            k8s_manager=k8s_manager,
            pod=pod,
            **params,
        )

    return logs, new_time


async def collect_agent_service_logs(
    k8s_manager: AsyncK8sManager, pod: V1Pod
) -> List[V1Log]:
    if not pod or not pod.spec.containers:
        return []
    container = pod.spec.containers[0]
    return await handle_container_logs(
        k8s_manager=k8s_manager,
        pod=pod,
        container_name=container.name,
        tail_lines=V1Logs._CHUNK_SIZE * 3,
    )


async def query_k8s_pod_logs(
    k8s_manager: AsyncK8sManager,
    pod: V1Pod,
    last_time: Optional[datetime.datetime],
    stream: bool = False,
) -> Tuple[List[V1Log], Optional[datetime.datetime]]:
    new_time = now()
    params = {}
    if last_time:
        since_seconds = (new_time - last_time).total_seconds() - 1
        params["since_seconds"] = int(since_seconds)
    if stream:
        params["tail_lines"] = V1Logs._CHUNK_SIZE

    logs = await handle_pod_logs(k8s_manager=k8s_manager, pod=pod, **params)

    if logs:
        last_time = logs[-1].timestamp
    return logs, last_time


async def get_op_pos_and_services(
    k8s_manager: AsyncK8sManager,
    run_uuid: str,
    run_kind: str,
):
    pods = await k8s_manager.list_pods(
        label_selector=k8s_manager.get_managed_by_polyaxon(run_uuid)
    )
    services = []
    if V1RunKind.has_service(run_kind):
        services = await k8s_manager.list_services(
            label_selector=k8s_manager.get_managed_by_polyaxon(run_uuid)
        )

    return pods, services


async def get_op_spec(
    k8s_manager: AsyncK8sManager,
    run_uuid: str,
    run_kind: str,
):
    pods, services = await get_op_pos_and_services(
        k8s_manager=k8s_manager,
        run_uuid=run_uuid,
        run_kind=run_kind,
    )
    pods_list = {}
    for pod in pods or []:
        pods_list[
            pod.metadata.name
        ] = k8s_manager.api_client.sanitize_for_serialization(pod)
    services_list = {}
    for service in services or []:
        services_list[
            service.metadata.name
        ] = k8s_manager.api_client.sanitize_for_serialization(service)
    data = {"pods": pods_list, "services": services_list}
    return data, pods, services


async def get_agent_pods_and_services(
    k8s_manager: AsyncK8sManager,
):
    pods = await k8s_manager.list_pods(
        label_selector=k8s_manager.get_core_polyaxon(),
    )
    services = await k8s_manager.list_services(
        label_selector=k8s_manager.get_core_polyaxon(),
    )
    return pods, services


async def get_agent_spec(
    k8s_manager: AsyncK8sManager,
):
    pods, services = await get_agent_pods_and_services(
        k8s_manager=k8s_manager,
    )
    pods_list = {}
    for pod in pods or []:
        pods_list[
            pod.metadata.name
        ] = k8s_manager.api_client.sanitize_for_serialization(pod)
    data = {"pods": pods_list}
    services_list = {}
    for service in services or []:
        services_list[
            service.metadata.name
        ] = k8s_manager.api_client.sanitize_for_serialization(service)
    data["services"] = services_list
    return data, pods, services



import logging
import os
import sys

from functools import wraps
from typing import List, Union

from polyaxon._env_vars.keys import ENV_KEYS_DEBUG, ENV_KEYS_LOG_LEVEL

logger = logging.getLogger("polyaxon.cli")


def configure_logger(verbose):
    # DO NOT MOVE OUTSIDE THE FUNCTION!
    from polyaxon import settings
    from polyaxon._plugins.sentry import set_raven_client

    if verbose or settings.CLIENT_CONFIG.debug or os.environ.get(ENV_KEYS_DEBUG, False):
        log_level = logging.DEBUG
        os.environ[ENV_KEYS_LOG_LEVEL] = "DEBUG"
        settings.CLIENT_CONFIG.debug = True
    else:
        if not settings.CLIENT_CONFIG.disable_errors_reporting:
            set_raven_client()
        log_level = (
            logging.DEBUG
            if os.environ.get(ENV_KEYS_LOG_LEVEL) in ["debug", "DEBUG"]
            else logging.INFO
        )
        if settings.CLIENT_CONFIG.log_level:
            try:
                log_level = logging.getLevelName(
                    settings.CLIENT_CONFIG.log_level.upper()
                )
            except Exception:  # noqa
                pass
    logging.basicConfig(format="%(message)s", level=log_level, stream=sys.stdout)


def clean_outputs(fn):
    """Decorator for CLI with Sentry client handling.
    see https://github.com/getsentry/sentry-python/issues/862#issuecomment-712697356
    """

    @wraps(fn)
    def clean_outputs_wrapper(*args, **kwargs):
        from polyaxon import settings

        cli_config = settings.CLI_CONFIG
        if cli_config and cli_config.log_handler and cli_config.log_handler.dsn:
            import sentry_sdk

            try:
                sentry_sdk.flush()
                return fn(*args, **kwargs)
            except Exception as e:
                sentry_sdk.capture_exception(e)
                sentry_sdk.flush()
                raise e
            finally:
                sentry_sdk.flush()
        else:
            return fn(*args, **kwargs)

    return clean_outputs_wrapper


def reconfigure_loggers(
    logger_names: List[str],
    log_level: Union[int, str] = logging.WARNING,
    disable: bool = True,
):
    for logger_name in logger_names:
        logging.getLogger(logger_name).setLevel(log_level)
        if disable:
            logging.getLogger(logger_name).disabled = True



from typing import List, Optional

from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_section_spec import V1SectionSpec


class V1DashboardSpec(BaseAllowSchemaModel):
    xaxis: Optional[str]
    smoothing: Optional[int]
    ignore_outliers: Optional[bool]
    sample_size: Optional[int]
    sections: Optional[List[V1SectionSpec]]



# Platform
ENV_KEYS_PLATFORM_CONFIG = "POLYAXON_PLATFORM_CONFIG"

# Context path for Polyaxon's config files
ENV_KEYS_HOME = "POLYAXON_HOME"

# General
ENV_KEYS_NO_CONFIG = "POLYAXON_NO_CONFIG"
ENV_KEYS_SERVICE = "POLYAXON_SERVICE"
ENV_KEYS_ENVIRONMENT = "POLYAXON_ENVIRONMENT"
ENV_KEYS_HEADER = "POLYAXON_HEADER"
ENV_KEYS_HEADER_SERVICE = "POLYAXON_HEADER_SERVICE"
ENV_KEYS_AUTHENTICATION_TYPE = "POLYAXON_AUTHENTICATION_TYPE"
ENV_KEYS_DEBUG = "POLYAXON_DEBUG"
ENV_KEYS_TIMEOUT = "POLYAXON_TIMEOUT"
ENV_KEYS_TRACKING_TIMEOUT = "POLYAXON_TRACKING_TIMEOUT"
ENV_KEYS_TIME_ZONE = "POLYAXON_TIME_ZONE"
ENV_KEYS_WATCH_INTERVAL = "POLYAXON_WATCH_INTERVAL"
ENV_KEYS_INTERVAL = "POLYAXON_INTERVAL"
ENV_KEYS_LOG_LEVEL = "POLYAXON_LOG_LEVEL"
ENV_KEYS_K8S_NAMESPACE = "POLYAXON_K8S_NAMESPACE"
ENV_KEYS_K8S_NODE_NAME = "POLYAXON_K8S_NODE_NAME"
ENV_KEYS_K8S_POD_ID = "POLYAXON_K8S_POD_ID"
ENV_KEYS_PLATFORM_HOST = "POLYAXON_PLATFORM_HOST"
ENV_KEYS_HOST = "POLYAXON_HOST"
ENV_KEYS_API_VERSION = "POLYAXON_API_VERSION"
ENV_KEYS_VERIFY_SSL = "POLYAXON_VERIFY_SSL"
ENV_KEYS_SSL_CA_CERT = "POLYAXON_SSL_CA_CERT"
ENV_KEYS_CERT_FILE = "POLYAXON_CERT_FILE"
ENV_KEYS_KEY_FILE = "POLYAXON_KEY_FILE"
ENV_KEYS_ASSERT_HOSTNAME = "POLYAXON_ASSERT_HOSTNAME"
ENV_KEYS_CONNECTION_POOL_MAXSIZE = "POLYAXON_CONNECTION_POOL_MAXSIZE"
ENV_KEYS_LOGS_ROOT = "POLYAXON_LOG_ROOT"
ENV_KEYS_ARCHIVES_ROOT = "POLYAXON_ARCHIVES_ROOT"
ENV_KEYS_ARTIFACTS_ROOT = "POLYAXON_ARTIFACTS_ROOT"
ENV_KEYS_STATIC_ROOT = "POLYAXON_STATIC_ROOT"
ENV_KEYS_CONTEXT_ROOT = "POLYAXON_CONTEXT_ROOT"
ENV_KEYS_OFFLINE_ROOT = "POLYAXON_OFFLINE_ROOT"
ENV_KEYS_DISABLE_ERRORS_REPORTING = "POLYAXON_DISABLE_ERRORS_REPORTING"
ENV_KEYS_INTERVALS_COMPATIBILITY_CHECK = "POLYAXON_INTERVALS_COMPATIBILITY_CHECK"
ENV_KEYS_RETRIES = "POLYAXON_RETRIES"
ENV_KEYS_UPLOAD_SIZE = "POLYAXON_UPLOAD_SIZE"
ENV_KEYS_MAX_CONCURRENCY = "POLYAXON_MAX_CONCURRENCY"
ENV_KEYS_HAS_PROCESS_SIDECAR = "POLYAXON_HAS_PROCESS_SIDECAR"

# Secrets
ENV_KEYS_SECRET_KEY = "POLYAXON_SECRET_KEY"  # noqa
ENV_KEYS_SECRET_INTERNAL_TOKEN = "POLYAXON_SECRET_INTERNAL_TOKEN"  # noqa
ENV_KEYS_AUTH_TOKEN = "POLYAXON_AUTH_TOKEN"
ENV_KEYS_AUTH_USERNAME = "POLYAXON_AUTH_USERNAME"
ENV_KEYS_CUSTOM_ERRORS_OPTIONS = "POLYAXON_CUSTOM_ERRORS_OPTIONS"

# Containers
ENV_KEYS_AGENT_INSTANCE = "POLYAXON_AGENT_INSTANCE"
ENV_KEYS_RUN_INSTANCE = "POLYAXON_RUN_INSTANCE"
ENV_KEYS_CONTAINER_ID = "POLYAXON_CONTAINER_ID"
ENV_KEYS_HASH_LENGTH = "POLYAXON_HASH_LENGTH"

# Flags
ENV_KEYS_K8S_IN_CLUSTER = "POLYAXON_K8S_IN_CLUSTER"
ENV_KEYS_IS_MANAGED = "POLYAXON_IS_MANAGED"
ENV_KEYS_IS_OFFLINE = "POLYAXON_IS_OFFLINE"
ENV_KEYS_NO_OP = "POLYAXON_NO_OP"
ENV_KEYS_NO_API = "POLYAXON_NO_API"

# Registry
ENV_KEYS_USE_GIT_REGISTRY = "POLYAXON_USE_GIT_REGISTRY"
ENV_KEYS_PUBLIC_REGISTRY = "POLYAXON_PUBLIC_REGISTRY"

# Agent
ENV_KEYS_AGENT_SIDECAR = "POLYAXON_AGENT_SIDECAR"
ENV_KEYS_AGENT_INIT = "POLYAXON_AGENT_INIT"
ENV_KEYS_AGENT_CLEANER = "POLYAXON_AGENT_CLEANER"
ENV_KEYS_AGENT_NOTIFIER = "POLYAXON_AGENT_NOTIFIER"
ENV_KEYS_AGENT_DEFAULT_SCHEDULING = "POLYAXON_AGENT_DEFAULT_SCHEDULING"
ENV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS = "POLYAXON_AGENT_DEFAULT_IMAGE_PULL_SECRETS"
ENV_KEYS_AGENT_ARTIFACTS_STORE = "POLYAXON_AGENT_ARTIFACTS_STORE"
ENV_KEYS_AGENT_CONNECTIONS = "POLYAXON_AGENT_CONNECTIONS"
ENV_KEYS_SET_AGENT = "POLYAXON_SET_AGENT"
ENV_KEYS_K8S_APP_SECRET_NAME = "POLYAXON_K8S_APP_SECRET_NAME"  # noqa
ENV_KEYS_WATCH_CLUSTER = "POLYAXON_WATCH_CLUSTER"
ENV_KEYS_SINGLE_NAMESPACE = "POLYAXON_SINGLE_NAMESPACE"
ENV_KEYS_ADDITIONAL_NAMESPACES = "POLYAXON_ADDITIONAL_NAMESPACES"
ENV_KEYS_AGENT_SECRET_NAME = "POLYAXON_AGENT_SECRET_NAME"  # noqa
ENV_KEYS_AGENT_RUNS_SA = "POLYAXON_AGENT_RUNS_SA"
ENV_KEYS_AGENT_ENABLE_HEALTH_CHECKS = "POLYAXON_AGENT_ENABLE_HEALTH_CHECKS"
ENV_KEYS_AGENT_EXECUTOR_REFRESH_INTERVAL = "POLYAXON_AGENT_EXECUTOR_REFRESH_INTERVAL"
ENV_KEYS_AGENT_IS_REPLICA = "POLYAXON_AGENT_IS_REPLICA"
ENV_KEYS_AGENT_USE_PROXY_ENV_VARS_IN_OPS = "POLYAXON_AGENT_USE_PROXY_ENV_VARS_IN_OPS"

# Connections
ENV_KEYS_COLLECT_ARTIFACTS = "POLYAXON_COLLECT_ARTIFACTS"
ENV_KEYS_COLLECT_RESOURCES = "POLYAXON_COLLECT_RESOURCES"
ENV_KEYS_ARTIFACTS_STORE_NAME = "POLYAXON_ARTIFACTS_STORE_NAME"
ENV_KEYS_GIT_CREDENTIALS = "POLYAXON_GIT_CREDENTIALS"
ENV_KEYS_GIT_CREDENTIALS_STORE = "POLYAXON_GIT_CREDENTIALS_STORE"
ENV_KEYS_SSH_PATH = "POLYAXON_SSH_PATH"
ENV_KEYS_SSH_PRIVATE_KEY = "POLYAXON_SSH_PRIVATE_KEY"

# Ops
ENV_KEYS_ISTIO_ENABLED = "POLYAXON_ISTIO_ENABLED"
ENV_KEYS_TFJOB_ENABLED = "POLYAXON_TFJOB_ENABLED"
ENV_KEYS_PYTORCH_JOB_ENABLED = "POLYAXON_PYTORCH_JOB_ENABLED"
ENV_KEYS_PADDLE_JOB_ENABLED = "POLYAXON_PADDLE_JOB_ENABLED"
ENV_KEYS_MXJOB_ENABLED = "POLYAXON_MXJOB_ENABLED"
ENV_KEYS_XGBOOST_JOB_ENABLED = "POLYAXON_XGBOOST_JOB_ENABLED"
ENV_KEYS_MPIJOB_ENABLED = "POLYAXON_MPIJOB_ENABLED"
ENV_KEYS_RAYJOB_ENABLED = "POLYAXON_RAYJOB_ENABLED"
ENV_KEYS_DASKJOB_ENABLED = "POLYAXON_DASKJOB_ENABLED"

# Sandbox
ENV_KEYS_SANDBOX_PORT = "POLYAXON_SANDBOX_PORT"
ENV_KEYS_SANDBOX_HOST = "POLYAXON_SANDBOX_HOST"
ENV_KEYS_SANDBOX_DEBUG = "POLYAXON_SANDBOX_DEBUG"
ENV_KEYS_SANDBOX_SSL_ENABLED = "POLYAXON_SANDBOX_SSL_ENABLED"
ENV_KEYS_SANDBOX_WORKERS = "POLYAXON_SANDBOX_WORKERS"
ENV_KEYS_SANDBOX_PER_CORE = "POLYAXON_SANDBOX_PER_CORE"
ENV_KEYS_SANDBOX_ROOT = "POLYAXON_SANDBOX_ROOT"
ENV_KEYS_SERVICE_MODE = "POLYAXON_SERVICE_MODE"

# Proxies
ENV_KEYS_PROXY_NAMESPACES = "POLYAXON_PROXY_NAMESPACES"
ENV_KEYS_PROXY_GATEWAY_PORT = "POLYAXON_PROXY_GATEWAY_PORT"
ENV_KEYS_PROXY_GATEWAY_TARGET_PORT = "POLYAXON_PROXY_GATEWAY_TARGET_PORT"
ENV_KEYS_PROXY_GATEWAY_HOST = "POLYAXON_PROXY_GATEWAY_HOST"
ENV_KEYS_PROXY_GATEWAY_CONCURRENCY = "POLYAXON_PROXY_GATEWAY_CONCURRENCY"
ENV_KEYS_PROXY_GATEWAY_PER_CORE = "POLYAXON_PROXY_GATEWAY_PER_CORE"
ENV_KEYS_PROXY_STREAMS_PORT = "POLYAXON_PROXY_STREAMS_PORT"
ENV_KEYS_PROXY_STREAMS_TARGET_PORT = "POLYAXON_PROXY_STREAMS_TARGET_PORT"
ENV_KEYS_PROXY_STREAMS_HOST = "POLYAXON_PROXY_STREAMS_HOST"
ENV_KEYS_PROXY_API_PORT = "POLYAXON_PROXY_API_PORT"
ENV_KEYS_PROXY_API_TARGET_PORT = "POLYAXON_PROXY_API_TARGET_PORT"
ENV_KEYS_PROXY_API_HOST = "POLYAXON_PROXY_API_HOST"
ENV_KEYS_PROXY_LOCAL_PORT = "POLYAXON_PROXY_LOCAL_PORT"
ENV_KEYS_PROXY_API_USE_RESOLVER = "POLYAXON_PROXY_API_USE_RESOLVER"
ENV_KEYS_PROXY_SSL_PATH = "POLYAXON_PROXY_SSL_PATH"
ENV_KEYS_PROXY_SSL_ENABLED = "POLYAXON_PROXY_SSL_ENABLED"
ENV_KEYS_PROXY_AUTH_ENABLED = "POLYAXON_PROXY_AUTH_ENABLED"
ENV_KEYS_PROXY_AUTH_EXTERNAL = "POLYAXON_PROXY_AUTH_EXTERNAL"
ENV_KEYS_PROXY_AUTH_USE_RESOLVER = "POLYAXON_PROXY_AUTH_USE_RESOLVER"
ENV_KEYS_PROXY_HAS_FORWARD_PROXY = "POLYAXON_PROXY_HAS_FORWARD_PROXY"
ENV_KEYS_PROXY_FORWARD_PROXY_PORT = "POLYAXON_PROXY_FORWARD_PROXY_PORT"
ENV_KEYS_PROXY_FORWARD_PROXY_HOST = "POLYAXON_PROXY_FORWARD_PROXY_HOST"
ENV_KEYS_PROXY_FORWARD_PROXY_KIND = "POLYAXON_PROXY_FORWARD_PROXY_KIND"
ENV_KEYS_UI_IN_SANDBOX = "POLYAXON_UI_IN_SANDBOX"
ENV_KEYS_UI_ADMIN_ENABLED = "POLYAXON_UI_ADMIN_ENABLED"
ENV_KEYS_UI_ASSETS_VERSION = "POLYAXON_UI_ASSETS_VERSION"
ENV_KEYS_UI_ENABLED = "POLYAXON_UI_ENABLED"
ENV_KEYS_UI_OFFLINE = "POLYAXON_UI_OFFLINE"
ENV_KEYS_UI_BASE_URL = "POLYAXON_UI_BASE_URL"
ENV_KEYS_STATIC_URL = "POLYAXON_STATIC_URL"
ENV_KEYS_DNS_USE_RESOLVER = "POLYAXON_DNS_USE_RESOLVER"
ENV_KEYS_DNS_CUSTOM_CLUSTER = "POLYAXON_DNS_CUSTOM_CLUSTER"
ENV_KEYS_DNS_BACKEND = "POLYAXON_DNS_BACKEND"
ENV_KEYS_DNS_PREFIX = "POLYAXON_DNS_PREFIX"
ENV_KEYS_NGINX_TIMEOUT = "POLYAXON_NGINX_TIMEOUT"
ENV_KEYS_NGINX_INDENT_CHAR = "POLYAXON_NGINX_INDENT_CHAR"
ENV_KEYS_NGINX_INDENT_WIDTH = "POLYAXON_NGINX_INDENT_WIDTH"



from typing import Dict, List, Optional, Union

from polyaxon._flow.io.io import V1IO
from polyaxon._flow.joins import V1Join
from polyaxon._flow.matrix import V1Mapping, V1Matrix
from polyaxon._flow.params.params import ParamSpec, V1Param
from polyaxon.exceptions import PolyaxonValidationError


def validate_params(
    params: Dict[str, Union[Dict, V1Param]],
    inputs: List[V1IO],
    outputs: List[V1IO],
    contexts: Optional[List[V1IO]] = None,
    matrix: Optional[Union[Dict, V1Matrix]] = None,
    joins: List[V1Join] = None,
    context: Optional[Dict] = None,
    is_template: bool = True,
    check_all_refs: bool = False,
    extra_info: Optional[str] = None,
    parse_values: bool = False,
) -> List[ParamSpec]:
    """
    Validates Params given inputs, and an optional context.

    Params can be:
     * plain values: we check them against the inputs types
     * job/experiment references: We postpone to server side validation.
     * ops reference: in that case a context must be provided to validate that the reference exists.
        and types are correct.
    """
    contexts_by_keys = {k.name: k for k in contexts or []}

    def parse_param(k, v) -> V1Param:
        if not isinstance(v, V1Param):
            v = V1Param.read(v, config_type=".yaml") if v else V1Param()
        if v and k in contexts_by_keys:
            v.context_only = True
            if v.to_init is None:
                v.to_init = contexts_by_keys[k].to_init
            if v.to_env is None:
                v.to_env = contexts_by_keys[k].to_env
            if v.connection is None:
                v.connection = contexts_by_keys[k].connection
        return v

    def validate_matrix(io: V1IO) -> bool:
        if isinstance(matrix, V1Mapping):
            return matrix.has_key(io.name)  # noqa
        elif io.name in matrix.params:
            matrix.params[io.name].validate_io(io)
            return True
        elif hasattr(matrix, "resource") and io.name == matrix.resource.name:
            return True
        return False

    params = params or {}
    params = {k: parse_param(k, params[k]) for k in params}
    # Extend with params coming from joins
    params.update(
        {k: v for j in (joins or []) for k, v in j.params.items() if k not in params}
    )

    if requires_params(inputs, outputs):
        if not is_template and not params and not matrix:
            message = (
                "The Polyaxonfile has non optional inputs/outputs, "
                "you need to pass valid params"
            )
            if extra_info:
                message += " Please check: {}".format(extra_info)

            raise PolyaxonValidationError(message)
    elif not accepts_params(inputs, outputs) and params:
        extra_params = set(params.keys()) - {
            p for p in params if params[p].context_only
        }
        if extra_params:
            message = "Received unexpected params `{}`".format(extra_params)
            if extra_info:
                message += " Please check: {}".format(extra_info)
            raise PolyaxonValidationError(message)

    matrix = matrix or {}
    inputs = inputs or []
    outputs = outputs or []

    processed_params = []
    validated_params = []

    for inp in inputs:
        if inp.name in params:
            param_value = params[inp.name]
            param_spec = param_value.get_spec(
                name=inp.name,
                iotype=inp.type,
                is_flag=inp.is_flag,
                is_list=inp.is_list,
                is_context=False,
                arg_format=inp.arg_format,
            )
            if param_spec.param.is_ref or param_spec.param.is_template_ref:
                param_spec.validate_ref(context, is_template, check_all_refs)
            else:
                parsed_value = inp.validate_value(param_value.value)
                if parse_values:
                    param_spec.param.value = parsed_value
                    if not param_spec.param.connection and param_value.connection:
                        param_spec.param.connection = inp.connection
                    if not param_spec.param.to_init and param_value.to_init:
                        param_spec.param.to_init = inp.to_init
                    if not param_spec.param.to_env and param_value.to_env:
                        param_spec.param.to_env = inp.to_env
            validated_params.append(param_spec)
            if not param_spec.param.context_only:
                processed_params.append(inp.name)
        elif matrix and validate_matrix(inp):
            pass
        elif not (inp.is_optional or inp.delay_validation or is_template):
            message = "Input {} is required, no param was passed.".format(inp.name)
            if extra_info:
                message += " Please check: {}".format(extra_info)
            raise PolyaxonValidationError(message)
        else:
            validated_params.append(
                ParamSpec(
                    name=inp.name,
                    param=V1Param(
                        value=inp.value,
                        connection=inp.connection,
                        to_init=inp.to_init,
                        to_env=inp.to_env,
                    ),
                    type=inp.type,
                    is_flag=inp.is_flag,
                    is_list=inp.is_list,
                    is_context=False,
                    is_requested=True,
                    arg_format=inp.arg_format,
                )
            )

    for out in outputs:
        if out.name in params:
            param_value = params[out.name]
            param_spec = param_value.get_spec(
                name=out.name,
                iotype=out.type,
                is_flag=out.is_flag,
                is_list=out.is_list,
                is_context=False,
                arg_format=out.arg_format,
            )
            if param_spec.param.is_ref or param_spec.param.is_template_ref:
                param_spec.validate_ref(
                    None, is_template=False, check_all_refs=check_all_refs
                )
            else:  # Plain value
                parsed_value = out.validate_value(param_value.value)
                if parse_values:
                    param_spec.param.value = parsed_value
                    if not param_spec.param.connection and param_value.connection:
                        param_spec.param.connection = out.connection
                    if not param_spec.param.to_init and param_value.to_init:
                        param_spec.param.to_init = out.to_init
                    if not param_spec.param.to_env and param_value.to_env:
                        param_spec.param.to_env = out.to_env
            validated_params.append(param_spec)
            if not param_spec.param.context_only:
                processed_params.append(out.name)
        # No validation for outputs we assume that the op might populate a context or send a metric
        elif out.delay_validation is not False:
            validated_params.append(
                ParamSpec(
                    name=out.name,
                    param=V1Param(
                        value=out.value,
                        connection=out.connection,
                        to_init=out.to_init,
                        to_env=out.to_env,
                    ),
                    type=out.type,
                    is_flag=out.is_flag,
                    is_list=out.is_list,
                    is_context=False,
                    is_requested=True,
                    arg_format=out.arg_format,
                )
            )
    extra_params = set(params.keys()) - set(processed_params)
    context_params = {p for p in params if params[p].context_only}
    extra_invalid_params = extra_params - context_params
    if extra_invalid_params:
        message = "Received unexpected params `{}`".format(extra_invalid_params)
        if extra_info:
            message += " Please check: {}".format(extra_info)
        raise PolyaxonValidationError(message)

    # Add all (extra) context params that were not processed during the IO check
    for p in contexts_by_keys:
        if p in extra_params:
            param_value = params[p]
            param_spec = param_value.get_spec(
                name=p,
                iotype=None,
                is_flag=None,
                is_list=None,
                is_context=True,
                arg_format=None,
            )
            validated_params.append(param_spec)
        else:
            context_io = contexts_by_keys[p]
            validated_params.append(
                ParamSpec(
                    name=p,
                    param=V1Param(
                        value=context_io.value,
                        connection=context_io.connection,
                        to_init=context_io.to_init,
                        to_env=context_io.to_env,
                    ),
                    type=None,
                    is_flag=None,
                    is_list=None,
                    is_context=True,
                    is_requested=True,
                    arg_format=None,
                )
            )
    return validated_params


def requires_params(inputs: List[V1IO], outputs: List[V1IO]):
    if not inputs and not outputs:
        return False

    def parse_io(ios: List[V1IO], delay_validation: bool):
        if not ios:
            return False
        for i in ios:
            delay_validation = (
                i.delay_validation
                if i.delay_validation is not None
                else delay_validation
            )
            if not i.is_optional and not delay_validation:
                return True
        return False

    if parse_io(inputs, False):
        return True

    if parse_io(outputs, True):
        return True

    return False


def accepts_params(inputs: List[V1IO], outputs: List[V1IO]):
    return bool(inputs or outputs)


def get_upstream_op_params_by_names(params: Dict[str, V1Param]):
    upstream = {}

    if not params:
        return upstream

    for param in params:
        param_ref = params[param].get_spec(
            name=param,
            iotype=None,
            is_flag=None,
            is_list=None,
            is_context=None,
            arg_format=None,
        )
        if param_ref and param_ref.param.is_ops_ref:
            if param_ref.param.entity_ref in upstream:
                upstream[param_ref.param.entity_ref].append(param_ref)
            else:
                upstream[param_ref.param.entity_ref] = [param_ref]

    return upstream


def get_upstream_run_params_by_names(params: Dict[str, V1Param]):
    upstream = {}

    if not params:
        return upstream

    for param in params:
        param_ref = params[param].get_spec(
            name=param,
            iotype=None,
            is_flag=None,
            is_list=None,
            is_context=None,
            arg_format=None,
        )
        if param_ref and param_ref.param.is_runs_ref:
            if param_ref.param.entity_ref in upstream:
                upstream[param_ref.param.entity_ref].append(param_ref)
            else:
                upstream[param_ref.param.entity_ref] = [param_ref]

    return upstream


def get_dag_params_by_names(params: Dict[str, V1Param]):
    upstream = {}

    if not params:
        return upstream

    for param in params:
        param_ref = params[param].get_spec(
            name=param,
            iotype=None,
            is_flag=None,
            is_list=None,
            is_context=None,
            arg_format=None,
        )
        if param_ref and param_ref.param.is_dag_ref:
            if param_ref.param.entity_ref in upstream:
                upstream[param_ref.param.entity_ref].append(param_ref)
            else:
                upstream[param_ref.param.entity_ref] = [param_ref]

    return upstream



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_agent import V1Agent


class V1ListAgentsResponse(BaseAllowSchemaModel):
    count: Optional[int]
    results: Optional[List[V1Agent]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



import asyncio
import os

from clipped.utils.tz import now
from kubernetes_asyncio.client.rest import ApiException

from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.getters import get_run_info
from polyaxon._env_vars.keys import ENV_KEYS_K8S_POD_ID
from polyaxon._fs.fs import (
    close_fs,
    get_artifacts_connection,
    get_async_fs_from_connection,
)
from polyaxon._fs.watcher import FSWatcher
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from polyaxon._sidecar.container.intervals import get_sync_interval
from polyaxon._sidecar.container.monitors import sync_artifacts, sync_logs, sync_spec
from polyaxon._sidecar.ignore import CONTAINER_IGNORE_FOLDERS
from polyaxon.client import RunClient
from polyaxon.exceptions import PolyaxonClientException, PolyaxonContainerException
from polyaxon.logger import logger
from polyaxon.settings import CLIENT_CONFIG


async def start_sidecar(
    container_id: str,
    sleep_interval: int,
    sync_interval: int,
    monitor_outputs: bool,
    monitor_logs: bool,
    monitor_spec: bool,
):
    sync_interval = get_sync_interval(
        interval=sync_interval, sleep_interval=sleep_interval
    )
    try:
        pod_id = os.environ[ENV_KEYS_K8S_POD_ID]
    except KeyError as e:
        raise PolyaxonContainerException(
            "Please make sure that this job has been "
            "started by Polyaxon with all required context."
        ) from e

    try:
        owner, project, run_uuid = get_run_info()
    except PolyaxonClientException as e:
        raise PolyaxonContainerException(e)

    client = RunClient(owner=owner, project=project, run_uuid=run_uuid)
    k8s_manager = AsyncK8sManager(namespace=CLIENT_CONFIG.namespace, in_cluster=True)
    await k8s_manager.setup()
    pod = await k8s_manager.get_pod(pod_id, reraise=True)
    connection = get_artifacts_connection()
    fs = await get_async_fs_from_connection(connection=connection)
    if os.path.exists(ctx_paths.CONTEXT_MOUNT_FILE_WATCHER):
        fw = FSWatcher.read(ctx_paths.CONTEXT_MOUNT_FILE_WATCHER)
    else:
        fw = FSWatcher()

    retry = 0
    is_running = True
    counter = 0
    state = {
        "last_artifacts_check": None,
        "last_logs_check": None,
    }

    async def monitor():
        if monitor_spec and pod.metadata.annotations:
            await sync_spec(
                k8s_manager=k8s_manager,
                run_uuid=run_uuid,
                run_kind=pod.metadata.annotations.get("operation.polyaxon.com/kind"),
            )
        if monitor_logs:
            await sync_logs(
                run_uuid=run_uuid,
                k8s_manager=k8s_manager,
                pod=pod,
                last_time=None,
                stream=True,
                is_running=is_running,
            )
        if monitor_outputs:
            last_check = state["last_artifacts_check"]
            try:
                await sync_artifacts(
                    fs=fs,
                    fw=fw,
                    store_path=connection.store_path,
                    run_uuid=run_uuid,
                    exclude=CONTAINER_IGNORE_FOLDERS,
                )
            except Exception as e:
                logger.debug(
                    "An error occurred while syncing artifacts, Exception %s" % repr(e)
                )
            try:
                client.sync_events_summaries(
                    last_check=last_check,
                    events_path=ctx_paths.CONTEXT_MOUNT_RUN_EVENTS_FORMAT.format(
                        run_uuid
                    ),
                )
                client.sync_system_events_summaries(
                    last_check=last_check,
                    events_path=ctx_paths.CONTEXT_MOUNT_RUN_SYSTEM_RESOURCES_EVENTS_FORMAT.format(
                        run_uuid
                    ),
                )
                update_last_check = True
            except Exception as e:
                logger.debug(
                    "An error occurred while syncing events summaries, "
                    "Exception %s" % repr(e)
                )
                update_last_check = False
            if update_last_check:
                state["last_artifacts_check"] = now()

    while is_running and retry <= 3:
        await asyncio.sleep(sleep_interval)
        if retry:
            await asyncio.sleep(retry * 2)
        try:
            is_running = await k8s_manager.is_pod_running(pod_id, container_id)
        except ApiException as e:
            retry += 1
            logger.info("Exception %s" % repr(e))
            logger.info("Sleeping ...")
            continue

        logger.debug("Syncing ...")
        if is_running:
            retry = 0

        counter += 1
        if counter == sync_interval:
            counter = 0
            try:
                await monitor()
            except Exception as e:
                logger.warning("Polyaxon sidecar error: %s" % repr(e))

    await monitor()
    logger.info("Cleaning non main containers")
    if k8s_manager:
        logger.info("Cleaning k8s manager")
        await k8s_manager.close()

    logger.info("Cleaning fs connection")
    await close_fs(fs)
    # Ensures that the monitors are closed
    await asyncio.sleep(1)



from typing import Any, Dict, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1AgentReconcileBodyRequest(BaseAllowSchemaModel):
    """
    V1AgentReconcileBodyRequest
    """

    owner: Optional[StrictStr] = None
    uuid: Optional[StrictStr] = None
    reconcile: Optional[Dict[str, Any]] = None



from typing import List, Optional, Union

from clipped.compact.pydantic import StrictStr
from clipped.types.ref_or_obj import BoolOrRef, RefField

from polyaxon._schemas.base import BaseSchemaModel


class V1Template(BaseSchemaModel):
    """A template is way for users to define
    specifications (components/operations) and signal to the CLI/API
    that they are not executable without modification.

    Oftentimes, a user might create a component or an operation that depends on a connection,
    if the user decides to share this specification they will have to face two choices:
     * Remove the connection, other users will be wondering why the specification does not work.
     * Leave the connection, other users will only know about
       the error after they execute the specification.

    By setting the template section, users can signal that this is not executable,
    and if a user runs the specification accidentally,
    they will be shown an error that gets augmented by instruction from the template's description,
    and optionally the fields that need to be modifier.

    Args:
        enabled: bool, optional
        description: str, optional
        fields: List[str], optional

    ## YAML usage

    ```yaml
    >>> template:
    >>>   enabled: true
    >>>   description:
    >>>   fields:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Template
    >>> template = V1Template(
    >>>     enabled=True,
    >>>     description="This operation cannot be executed and requires proper annotations.",
    >>>     fields=["environment.annotations", "environment.labels"],
    >>> )
    ```

    ## Fields

    ### enabled

    A flag to enable the template behavior.

    ```yaml
    >>> template:
    >>>   enabled: true
    ```

    ### description

    An optional description about the template.
    The description is also helpful to extend the error raised by the CLI/API.

    ```yaml
    >>> template:
    >>>   description: "some more information about why this is a template and how to execute it"
    ```

    ### fields

    An optional list of field paths that need to be customized to run the specification.

    ```yaml
    >>> template:
    >>>   fields: ["environment.annotations", "environment.labels"]
    ```
    """

    _IDENTIFIER = "template"

    enabled: Optional[BoolOrRef]
    description: Optional[StrictStr]
    fields: Optional[Union[List[StrictStr], RefField]]


class TemplateMixinConfig:
    def disable_template(self):
        if self.is_template():
            self.template.enabled = False

    def is_template(self):
        return self.template and self.template.enabled
